/*
 * Copyright (c) 1997 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * acltest.c: checks to see if a user is permitted to access a
 *            path, according to AFS acls or .htaccess files.
 *
 * this is a stupid hack. i know. shut up.
 *
 * See http://www-personal.umich.edu/~dugsong/krb4nsapi for more info.
 *
 * Copyright (c) 1997 Dug Song <dugsong@UMICH.EDU>
 * All rights reserved, all wrongs reversed.
 * Disclaimer: i claim dis
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>

#include <netinet/in.h>

#ifdef AFS
/* Standard AFS includes. */
#include <sys/ioccom.h>
#include <afs/param.h>
#include <afs/stds.h>
#include <afs/vice.h>
#include <afs/venus.h>
#include <afs/prs_fs.h>
#include <afs/afsint.h>
#include <afs/cellconfig.h>
#ifdef USE_PTSCD
#include "ptscd.h"
#endif /* USE_PTSCD */
#else /* ! AFS */
#define PRSFS_READ	1
#define PRSFS_LOOKUP	8
#endif /* AFS */

/* Some ACL definitions. */
#define MAX_ACLENTRIES 20
#define MAX_ACLNAME 100
#define MAX_ACLSIZE 2048

typedef struct Acl_entry_type {
  char name[MAX_ACLNAME];
  long rights;
} Acl_entry;
 
typedef struct Acl_type {
  unsigned int num_plus;
  unsigned int num_minus;
  Acl_entry plus[MAX_ACLENTRIES];
  Acl_entry minus[MAX_ACLENTRIES];
} Acl;

#ifdef AFS
int afs_get_acl(Acl *acl, const char *path)
{
  struct ViceIoctl acljunk;
  char *c, tmp[MAX_ACLSIZE];
  int i;
 
  if (strlen(path) > MAXPATHLEN) return -1;
  
  acljunk.out_size = sizeof(tmp);
  acljunk.in_size = 0;
  acljunk.out = tmp;
  
  if (pioctl(path, VIOCGETAL, &acljunk, 1)) return -1;
  
  c = tmp;
  sscanf(c, "%d", &acl->num_plus);
  c = strchr(c, '\n'); c++;
  sscanf(c, "%d", &acl->num_minus);
  c = strchr(c, '\n'); c++;
  
  for (i = 0; i < acl->num_plus; i++) {
    sscanf(c, "%100s %d", acl->plus[i].name, &acl->plus[i].rights);
    c = strchr(c, '\n'); c++;
  }
  for (i = 0; i < acl->num_minus ; i++) {
    sscanf(c, "%100s %d", acl->minus[i].name, &acl->minus[i].rights);
    c = strchr(c, '\n'); c++;
  }
  if ((acl->num_plus + acl->num_minus) > MAX_ACLENTRIES) return -1;
  
  return 0;
}
#endif /* AFS */

int htaccess_get_acl(Acl *acl, const char *htaccess)
{
  FILE *f;
  char buf[BUFSIZ], *cp;
  int valid_auth_type = 0;
  int valid_require_line = 0;
  static char *sep = " ,\t";
  
  acl->num_plus = acl->num_minus = 0;
    
  if ((f = fopen(htaccess, "r")) == NULL) return -1;
  while (fgets(buf, sizeof(buf), f)) {
    char *directive;
    char *what;
    char *who;
    
    /* Get rid of leading whitespace, skip comments
       and empty lines, and strip newlines. */
    for (cp = buf; *cp == ' ' || *cp == '\t'; cp++) ;
    if (*cp == '#' || *cp == '\n' || !*cp) continue;
    if (cp = strchr(buf, '\n')) *cp = 0;
    
    if (!(directive = strtok(buf, sep))) continue;
    
    if (strncasecmp(directive, "AuthType", 8) == 0) {
      /* Check for KerberosV4 auth type. */
      if ((what = strtok(NULL, sep)) != NULL &&
	  strncasecmp(what, "KerberosV4", 10) == 0)
	valid_auth_type = 1;
    }
    else if (strncasecmp(directive, "require", 7) == 0) {
      /* Check for valid require option and arguments. */
      if (!(what = strtok(NULL, sep))) continue;
      
      if (strncasecmp(what, "valid-user", 10) == 0) {
	strcpy(acl->plus[acl->num_plus].name, "system:anyuser");
	acl->plus[acl->num_plus++].rights = PRSFS_READ|PRSFS_LOOKUP;
	strcpy(acl->plus[acl->num_plus++].name, "system:authuser");
	acl->plus[acl->num_plus++].rights = PRSFS_READ|PRSFS_LOOKUP;
	valid_require_line = 1;
      }
      else if ((strncasecmp(what, "user", 4) == 0)
#ifdef AFS
	       || (strncasecmp(what, "group", 5) == 0)
#endif /* AFS */
	       )
	{
	  for (who = strtok(NULL, sep); who; who = strtok(NULL, sep)) {
	    strncpy(acl->plus[acl->num_plus].name, who, MAX_ACLNAME);
	    acl->plus[acl->num_plus++].rights = PRSFS_READ|PRSFS_LOOKUP;
	    valid_require_line = 1;
	  }
	}
    }
  }
  if (valid_auth_type && valid_require_line) return 0;
  return 1;
}

int user_in_acl(const char *user, Acl_entry *acl_list, int num_entries)
{
  int i, flag;

#if defined(AFS) && !defined(USE_PTSCD)
  if (pr_Initialize(0, AFSCONF_CLIENTNAME, 0)) return 0;
#endif /* AFS && !USE_PTSCD */
  for (i = 0; i < num_entries; i++) {
    if ((
#ifdef USE_PTSCD
	 pts_check(user, acl_list[i].name) ||
#elif AFS
	 (!pr_IsAMemberOf(user, acl_list[i].name, &flag) && flag) ||
#endif /* USE_PTSCD || AFS */
	 !strncmp(acl_list[i].name, "system:anyuser", 14) ||
	 !strncmp(acl_list[i].name, "system:authuser", 15) ||
	 !strncmp(acl_list[i].name, user, strlen(user))) &&
	(acl_list[i].rights & PRSFS_READ) &&
	(acl_list[i].rights & PRSFS_LOOKUP))
      return 1;
  }
  return 0;
}

void main(int argc, char *argv[])
{
  Acl acl;
  struct stat st;
  char *user = argv[1];
  char *path = argv[2];
  char *htaccess;
  int denied;
  
  if (argc != 3 || !user || !path) {
    fprintf(stderr, "Usage: %s [username] [pathname]\n", argv[0]);
    exit(2);
  }
  if (stat(path, &st) == -1) {
    perror("stat");
    exit(2);
  }
  htaccess = strdup(path);
  if (!S_ISDIR(st.st_mode)) *(strrchr(htaccess, '/')+1) = '\0';
  htaccess = (char *)realloc(htaccess, strlen(htaccess)+10+1);
  strcpy(htaccess+strlen(htaccess), "/.htaccess");

  /* Get ACLs from either .htaccess or AFS. .htaccess overrides. */
  if (htaccess_get_acl(&acl, htaccess) != 0) {
#ifdef AFS
    if (afs_get_acl(&acl, path) != 0) {
      fprintf(stderr, "Couldn't get ACLs from .htaccess or AFS!\n");
      exit(2);
    }
#else
    /* Return success if .htaccess isn't available. */
    exit(0);
#endif /* AFS */
  }
  denied = (!user_in_acl(user, acl.plus, acl.num_plus)
#ifdef AFS
	    || user_in_acl(user, acl.minus, acl.num_minus)
#endif /* AFS */
	    );
  
  fprintf(stderr, "User %s is %s access to %s\n", user,
	  denied ? "denied" : "allowed", argv[2]);
  
  exit(denied);
}
